"""
Script for adding dataset tasks to Seqio registry
This is done for all datasets sotred in DATASET_PATH folder
Script should be imported to mt3/tasks.py
"""
import os
import tensorflow as tf
from absl import logging

import functools
from mt3.datasets import InferEvalSplit, DatasetConfig
from mt3.tasks import add_transcription_task_to_registry
from mt3 import spectrograms, vocabularies, preprocessors

# Configuration
DATASET_PATH = "/cluster/home/fredrifw/TFRecords/"
FEATURES = {
    'id': tf.io.FixedLenFeature([], dtype=tf.string),
    'audio': tf.io.FixedLenFeature([], dtype=tf.string),
    'sequence': tf.io.FixedLenFeature([], dtype=tf.string),
    'instrument_sequences': tf.io.FixedLenSequenceFeature([], dtype=tf.string, allow_missing=True),
    'program_num': tf.io.FixedLenSequenceFeature([], dtype=tf.int64, allow_missing=True),
    'inst_names': tf.io.FixedLenSequenceFeature([], dtype=tf.string, allow_missing=True),
}

# Just use default spectrogram config.
SPECTROGRAM_CONFIG = spectrograms.SpectrogramConfig()

# Create vocabulary config for no velocity
VOCAB_CONFIG_NOVELOCITY = vocabularies.VocabularyConfig(num_velocity_bins=1)

def program_num_to_program(program_num: str) -> int:
    """
    Pipe from program_num to program to use in preprocessors.tokenize_example_with_program_lookup.
    Simply returns the same program number as in parameter (with some typechecks).
    Note: program_num input is str due to MT3 lambda constraints

    :param program_num: The program number in range [0, 127] (str)
    """
    # Typecheck for string
    if type(program_num) != str:
        raise TypeError(f"Program number '{program_num}' is not a string")

    program_num = int(program_num)

    # Check if program number is valid
    if 0 < program_num < 127:
        return program_num
    else:
        raise ValueError(f"Program number '{program_num}' is not in range [0, 127]")


datasets = [os.path.join(DATASET_PATH, x) for x in os.listdir(DATASET_PATH)]
for dataset in datasets:
    logging.info("Adding tasks for dataset stored in", dataset)

    # Split train partition and validation
    dataset_partition = []
    dataset_validation = None
    for partition in os.listdir(dataset):
        if 'validation' not in partition:
            #logging.info("-- Found partition", partition)
            dataset_partition.append(os.path.join(dataset, partition))
        else:
            #logging.info("-- Found validation partition", partition)
            if dataset_validation:
                logging.info("-- ERROR: Found several validation partitions")
            dataset_validation = os.path.join(dataset, partition)

    for partition in dataset_partition:
        splits = partition.split('/')
        partition_name = f"{splits[-2]}_{splits[-1].split('.')[0]}"

        paths = {'train': partition, 'validation': dataset_validation}

        config = DatasetConfig(
            name=partition_name,
            paths=paths,
            features=FEATURES,
            train_split="train",
            train_eval_split="validation",
            infer_eval_splits=[InferEvalSplit(name='validation', suffix='validation')]
        )

        # Add transcription task to registry
        add_transcription_task_to_registry(
            dataset_config=config,
            spectrogram_config=SPECTROGRAM_CONFIG,
            vocab_config=VOCAB_CONFIG_NOVELOCITY,
            tokenize_fn=functools.partial(
                preprocessors.tokenize_transcription_example,
                audio_is_samples=False,
                id_feature_key='id'),
            onsets_only=False,
            include_ties=True)

        #logging.info("Dataset task was added to registry as %s_notes_ties_vb1", partition_name)