import os
from absl import logging
from dataset_config import *
import tensorflow as tf
import pathlib

from Dataset.tfrecord_utils import TFRecordUtils

class TFRecordGenerator:
    """ Class for generation of a dataset for MIR tasks.
    Converts information of audio and MIDI into tf.train.Example,
    which are stored in a TFRecord file.
    """

    @staticmethod
    def generate_tfrecord(dataset, name, data_list):
        """ Generates a tfrecord from a list of data.

        :param dataset: name of dataset
        :param name: name of dataset partition, preferably in ["validation", "baseline", "e1", "e2", "e3"]
        :param data_list: a list of dictionaries with fields:
            {
                'id': name of the entry / audio file (str),
                'audio': full path to audio file (str),
                'sequence': full path to mixed midi file (str),
                'instrument_sequences': full paths to per instrument midi files (list(str))
                'program_num': program numbers, corresponding to index in instrument_sequences (list(int))
            }
        """
        if not os.path.exists(TFRECORD_LOCATION):
            os.mkdir(TFRECORD_LOCATION)
        ds_folder = os.path.join(TFRECORD_LOCATION, dataset)
        if not os.path.exists(ds_folder):
            os.mkdir(ds_folder)

        filename = f"{name}.tfrecord"
        file_path = os.path.join(ds_folder, filename)   # Create filename path

        serialized_examples = []
        for data in data_list:
            #if all([pathlib.Path(x).is_file() for x in  # Validate data
            #        [data["audio"], data["sequence"]] + data["instrument_sequences"]]):
            if all([pathlib.Path(x).is_file() for x in [data["audio"], data["sequence"]]]):
                serialized_example = TFRecordUtils.serialize_example(data)
                serialized_examples.append(serialized_example)
                logging.info("Added entry for id %s", data["id"])
            else:
                logging.warning("Could not add entry for id '%s'", data["id"])

        with tf.io.TFRecordWriter(file_path) as writer:
            for serialized_example in serialized_examples:
                writer.write(serialized_example)

        logging.info("A TFRecord with %d examples was stored in %s", len(serialized_examples), file_path)


