import json
import os
import itertools
import yaml
import shutil
from utils import _load_audio_file, _merge_audio, merge_midi
from config import *
from scipy.io import wavfile
from mido import MidiFile
from copy import deepcopy
from tqdm import tqdm
from time import time
from paramiko import SSHClient, AutoAddPolicy


"""
File contains functions to derive the Cerberus dataset from Slakh2100, as described in the MT3 paper. 
"""

def _get_existing_audio(slakh_folder, output_folder, sftp_client):
    existing_audio = []
    for pipeline in tqdm(os.listdir(slakh_folder)):
        file_exists = False
        for dir in sftp_client.listdir(output_folder):
            file_path = f"{output_folder}/{dir}/{pipeline}_0"
            try:
                sftp_client.lstat(file_path)
                file_exists = True
            except FileNotFoundError:
                pass
        if file_exists:
            existing_audio.append(pipeline)
    return existing_audio

def _init_sftp_client(remote, user, pwd):
    ssh_client = SSHClient()
    ssh_client.set_missing_host_key_policy(AutoAddPolicy())
    ssh_client.connect(hostname=remote, username=user, password=pwd)
    return ssh_client.open_sftp()

def derive_cerberus_from_slakh(slakh_folder, cerberus_folder, debug=False, overwrite_existing=False):
    """
    Derives the Cerberus4 dataset from the Slakh2100 dataset, as described in MT3.
    :param slakh_folder: The folder containing the Slakh dataset
    :param cerberus_folder: The folder where the Cerberus dataset will be stored
    :return: None
    """
    existing = _get_existing_audio("Dataset/slakh2100_flac_redux/train", "Datasets/Cerberus", _init_sftp_client(SSH_HOST, SSH_USER, SSH_PASSWORD))
    if debug:
        print(f"Skipping {len(existing)} slakh tracks. Already exists. ")
    # Each directory is its own track
    for directory in tqdm(sorted(os.listdir(slakh_folder))):
        if directory.startswith("Track") and directory not in existing:
            with open(f"{slakh_folder}/{directory}/metadata.yaml", 'r') as file:
                track_meta = yaml.safe_load(file)
                combinations = _get_cerberus_subsets(f"{slakh_folder}/{directory}/stems", track_meta)
                if len(combinations) == 0:
                    print(f"No valid combinations for track {directory}. Skipping...")
                # Pre load audio for efficiency as they will appear in more than one mix
                audio_stems = load_audio_stems_from_combinations(combinations, f"{slakh_folder}/{directory}")
                # For each of the combinations, save the combination to a new folder
                # The folder should also include a full-mix, the midis (both all and stems) + new metadata
                for n in range(len(combinations)):
                    _save_combination_as_cerberus_track(combinations[n], n, f"{slakh_folder}/{directory}", cerberus_folder, track_meta, audio_stems)
        else:
            if debug:
                print(f"Skipping {directory}, output files already exist")


def load_audio_stems_from_combinations(combinations, stem_folder):
    stems = set()
    for combination in combinations:
        for i in combination:
            stems.add(i)
    return_audio = {}
    for stem in stems:
        stem_path = f"{stem_folder}/stems/{stem}"
        return_audio[stem] = _load_audio_file(stem_path, SAMPLE_RATE)
    return return_audio


def _get_cerberus_subsets(slakh_stem_folder, metadata):
    """
    Gets the valid cerberus subsets from a slakh_track
    :param slakh_stem_folder: The folder containing the stems for the track
    :param metadata: The parsed yaml metadata for the track
    :return: Returns a dict where key is track and val is list of subsets that constitute cerberus tracks
    """
    # Retrieve the stems for each; Drums, Guitar, Bass & Piano
    all_stems = [[stem for stem in os.listdir(slakh_stem_folder) if
                  metadata["stems"][stem.split(".")[0]]["inst_class"] == "Drums"],
                 [stem for stem in os.listdir(slakh_stem_folder) if
                  metadata["stems"][stem.split(".")[0]]["inst_class"] == "Guitar"],
                 [stem for stem in os.listdir(slakh_stem_folder) if
                  metadata["stems"][stem.split(".")[0]]["inst_class"] == "Bass"],
                 [stem for stem in os.listdir(slakh_stem_folder) if
                  metadata["stems"][stem.split(".")[0]]["inst_class"] == "Piano"]]
    combinations = list(itertools.product(*all_stems))
    return combinations


def _save_combination_as_cerberus_track(combination, n, slakh_track_folder, cerberus_folder, track_meta, preload_audio=None):
    if not os.path.exists(cerberus_folder):
        os.mkdir(cerberus_folder)
    # First get the track name
    track = slakh_track_folder.split("/")[-1]
    # Create the cerberus track folder
    track_folder = f"{cerberus_folder}/{track}_{n}"
    if not os.path.exists(track_folder):
        os.mkdir(track_folder)
    # Save the stems and midis
    if not os.path.exists(f"{track_folder}/stems"):
        os.mkdir(f"{track_folder}/stems")
    if not os.path.exists(f"{track_folder}/MIDI"):
        os.mkdir(f"{track_folder}/MIDI")
    stem_audios = []
    midi_files = []
    for stem in tqdm(combination):
        # Copy the stems
        stem_path = f"{slakh_track_folder}/stems/{stem}"
        shutil.copyfile(stem_path, f"{track_folder}/stems/{stem}")
        # Load the stem for later merging
        if preload_audio:
            stem_audios.append(preload_audio[stem])
        else:
            stem_audios.append(_load_audio_file(stem_path, SAMPLE_RATE))
        # Copy the midi files
        midi_path = f"{slakh_track_folder}/MIDI/{stem.split('.')[0]}.mid"
        shutil.copyfile(midi_path, f"{track_folder}/MIDI/{stem.split('.')[0]}.mid")
        # Load the midi for later merging
        midi_files.append(MidiFile(midi_path))
    # Merge audio and save full mix
    audio = _merge_audio(stem_audios)
    wavfile.write(f"{track_folder}/mix.wav", SAMPLE_RATE, audio.transpose())
    # Merge midi and save as all source
    merged_midi = merge_midi(midi_files)
    merged_midi.save(f"{track_folder}/all_src.mid")
    # Copy the necessary metadata
    new_meta = construct_new_meta_yaml(combination, track_meta)
    file = open(f"{track_folder}/metadata.yaml", "w")
    yaml.dump(new_meta, file)
    file.close()



def construct_new_meta_yaml(stems, org_meta):
    """
    Constructs a new metadata-yaml file for the cerberus combination of stems.
    :param stems: The new combination of stems
    :param org_meta: Yaml metadata for the original Slakh track
    :return: The new metadata, containing only information for the new stems
    """
    new_meta = deepcopy(org_meta)
    for stem in org_meta["stems"]:
        if f"{stem}.flac" not in stems and f"{stem}.wav" not in stems:
            del new_meta["stems"][stem]
    return new_meta


if __name__ == '__main__':
    derive_cerberus_from_slakh("C:/Users/Fredrik/Documents/Skole/Masteroppgave/AMT/Dataset/slakh2100_flac_redux/train",
                               "C:/Users/Fredrik/Documents/Skole/Masteroppgave/AMT/Dataset/Cerberus/train", True)

