import os
from yaml import safe_load
import re
from immutabledict import immutabledict
import csv
from mido import MidiFile
from utils import merge_midi, _load_audio_file, _merge_audio
import itertools
from random import randint
from config import *




# To be used for validation and test
def parse_slakh_org(slakh_folder):
    songs = []
    for track in os.listdir(slakh_folder):
        if track.startswith("Track"):
            inst_numbers = []
            with open(f"{slakh_folder}/{track}/metadata.yaml", "r") as file:
                meta = safe_load(file)
                for stem in meta["stems"]:
                    inst_numbers.append(meta["stems"][stem]["program_num"])
            song = {"id": track, "audio": f"{slakh_folder}/{track}/mix.wav",
                    "sequence": f"{slakh_folder}/{track}/all_src.mid", "program_num": inst_numbers}
            songs.append(song)
    return songs


# To be used for training
def parse_slakh_org_with_subsets(slakh_folder, midi_subset_folder):
    """
    For each song
        Check if there already exists any subset mixes for this track (use or create new one with some prob)
        Now you can generate midi
        Then load audio and merge + convert to bytes: https://stackoverflow.com/questions/51300281/how-to-convert-numpy-array-to-bytes-object-without-save-audio-file-on-disk

    """
    songs = []
    for track in os.listdir(slakh_folder):
        if track.startswith("Track"):
            # Check if there already exists any subset-mixes:
            track_subset_midis = [midi for midi in os.listdir(midi_subset_folder) if
                                  re.match(f"^{track}(.*).mid$", midi)]
            existing_combos = [set(i.replace(".mid", "").split("_")[1:]) for i in track_subset_midis]
            available_stems = [stem.replace(".wav", "") for stem in os.listdir(f"{slakh_folder}/{track}/stems")]
            combinations = [set(combo) for combo in list(itertools.product(*[available_stems for _ in
                                                                             range(randint(4, len(available_stems)))]))
                            if len(combo) >= 4 and set(combo) not in existing_combos]
            # Retrieve track_stems
            while len(track_subset_midis) < 10:
                combination = combinations.pop()
                midi_name = f"{midi_subset_folder}/{track}_{'_'.join(combination)}.mid"
                midi = merge_midi([MidiFile(f"{slakh_folder}/{track}/MIDI/{stem}.mid") for stem in combination])
                midi.save(filename=midi_name)

            for midi in track_subset_midis:
                combo = midi.replace(".mid", "").split("_")[1:]
                # Get instrument numbers
                inst_numbers = []
                with open(f"{slakh_folder}/{track}/metadata.yaml", "r") as file:
                    meta = safe_load(file)
                    for stem in meta["stems"]:
                        if stem in combo:
                            inst_numbers.append(meta["stems"][stem]["program_num"])
                # Get individual audios and merge
                audios = [_load_audio_file(audio, SAMPLE_RATE) for audio in os.listdir(f"{slakh_folder}/{track}/stems")
                          if audio.replace(".wav", "") in combo]
                song = {"id": track, "audio": _merge_audio(audios),
                               "sequence": midi, "program_num": inst_numbers}
                songs.append(song)
    return songs


def parse_slakh_rendered(slakh_folder, org_folder, midi_subset_folder):
    """
    Parses a rendered slakh dataset.
    Since Slakh (as of MT3) is using different subsets of each track, this parser assumes that each rendered track file
    contains the used subset stems from the full mix. These are used by the parser to create a corresponding all source
    MIDI-file for the subset mix.
    :param slakh_folder: The folder containing the rendered slakh subsets
    :param org_folder: The folder containing the original Slakg dataset
    :param midi_subset_folder: The folder containing and/or to save all source midi subsets
    :return: The parsed dataset as input to a Tf-record parser
    """
    songs = []
    for track_folder in os.listdir(slakh_folder):
        if track.startswith("Track"):
            track = track_folder.split(".")[0]
            stems = track.split("_")[1:]
            inst_numbers = []
            with open(f"{org_folder}/{track}/metadata.yaml", "r") as file:
                meta = safe_load(file)
                for stem in meta["stems"]:
                    if stem in stems:
                        inst_numbers.append(meta["stems"][stem]["program_num"])
            # Get all-source.midi for subset mix
            midi_name = f"{midi_subset_folder}/{track}_{'_'.join(stems)}.mid"
            if not os.path.exists(midi_name):
                midi = merge_midi([MidiFile(f"{org_folder}/{track}/MIDI/{stem}.mid") for stem in stems])
                midi.save(filename=midi_name)
            song = {"id": track, "audio": f"{slakh_folder}/{track}",
                    "sequence": midi_name, "program_num": inst_numbers}
            songs.append(song)
    return songs


_URMP_INSTRUMENT_PROGRAMS = immutabledict({
    'vn': 40,  # violin
    'va': 41,  # viola
    'vc': 42,  # cello
    'db': 43,  # double bass
    'tpt': 56,  # trumpet
    'tbn': 57,  # trombone
    'tba': 58,  # tuba
    'hn': 60,  # French horn
    'sax': 64,  # saxophone
    'ob': 68,  # oboe
    'bn': 70,  # bassoon
    'cl': 71,  # clarinet
    'fl': 73  # flute
})

def parse_urmp_rendered(urmp_folder=os.getcwd(), org_folder=os.getcwd()):
    songs = []
    for track in os.listdir(urmp_folder):
        if track.endswith(".wav"):
            track_name = track.split(".")[0]
            midi_path = f"{org_folder}/{track_name}/{track_name}.mid"
            song = {"id": track_name, "audio": f"{urmp_folder}/{track}",
                    "sequence": midi_path}
            songs.append(song)
    return songs

def parse_urmp_org(urmp_folder=os.getcwd()):
    songs = []
    for track_folder in os.listdir(urmp_folder):
        if re.match("^[0-9]{2}_(.*)", track_folder):
            inst_names = track_folder.split("_")[2:]
            inst_numbers = []
            for i in inst_names:
                inst_numbers.append(_URMP_INSTRUMENT_PROGRAMS[i])
            # Find all-source midi and single midis
            midi_path = None
            note_sequences = [None for _ in range(len(inst_numbers))]
            for midi in os.listdir(f"{urmp_folder}/{track_folder}"):
                if re.match("^Sco(.*)(.mid)$", midi):
                    midi_path = midi
                elif re.match("^[0-9]{1}_(.*)(.mid)$", midi):
                    # We have a single instrument midi
                    note_sequences[int(midi.split("_")[0])] = f"{urmp_folder}/{track_folder}/{midi}"
            song = {"id": track_folder, "audio": f"{urmp_folder}/{track_folder}/AuMix_{track_folder}.wav",
                    "sequence": f"{urmp_folder}/{track_folder}/{track_folder}.mid", "program_num": inst_numbers,
                    "instrument_sequences": note_sequences}
            songs.append(song)
    return songs


def parse_guitarset_rendered(guitarset_folder=os.getcwd(), org_folder=os.getcwd()):
    songs = []
    for track in os.listdir(guitarset_folder):
        if track.endswith(".wav"):
            track_name = track.split(".")[0]
            inst_numbers = [24]  # Always 24 for guitarset
            # Find source midi
            midi_path = None
            for midi in os.listdir(f"{org_folder}/{track_name}"):
                if re.match(".*(.mid|.midi)$", midi):
                    midi_path = midi
                    break
            song = {"id": track_name, "audio": f"{guitarset_folder}/{track}",
                    "sequence": f"{org_folder}/{track_name}/{midi_path}", "program_num": inst_numbers,
                    "instrument_sequences": [f"{org_folder}/{track_name}/{midi_path}"]}
            songs.append(song)
    return songs


def parse_guitarset_org(guitarset_folder=os.getcwd()):
    songs = []
    for track in os.listdir(guitarset_folder):
        if re.match("^[0-9]{2}(.*)(comp|solo)$", track):
            inst_numbers = [24]
            # Find source midi
            midi_path = None
            for midi in os.listdir(f"{guitarset_folder}/{track}"):
                if re.match(".*(.mid|.midi)$", midi):
                    midi_path = midi
                    break
            song = {"id": track, "audio": f"{guitarset_folder}/{track}/{track}_hex.wav",
                    "sequence": f"{guitarset_folder}/{track}/{midi_path}", "program_num": inst_numbers,
                    "instrument_sequences": [f"{guitarset_folder}/{track}/{midi_path}"]}
            songs.append(song)
    return songs


def parse_musicnet_rendered(musicnet_folder=os.getcwd(), org_folder=os.getcwd()):
    songs = []
    skipped = []
    for track in os.listdir(musicnet_folder):
        if track.endswith(".wav"):
            track_name = track.split(".")[0]
            midi_path = f"{org_folder}/{track_name}/{track_name}.mid"
            # Check if MIDI path exists to see if item is train split
            if os.path.exists(midi_path):
                song = {"id": track_name, "audio": f"{musicnet_folder}/{track}",
                        "sequence": midi_path}
                songs.append(song)
            else:
                skipped.append(track)
    print("Skipped: ", skipped)
    return songs

def parse_musicnet_org(musicnet_folder=os.getcwd()):
    songs = []
    for track_folder in os.listdir(musicnet_folder):
        if re.match("^[0-9]{4}$", track_folder):
            midi_path = f"{musicnet_folder}/{track_folder}/{track_folder}.mid"
            if os.path.exists(midi_path):
                song = {"id": track_folder, "audio": f"{musicnet_folder}/{track_folder}/{track_folder}.wav",
                        "sequence": midi_path}
                songs.append(song)
            else:
                print(f"{track_folder} does not have MIDI")
    return songs


def parse_maestro_rendered(maestro_folder, org_folder):
    file = open("MAESTRO_SUBSET.txt", "r")
    lines = file.read().splitlines()
    file.close()

    songs = []
    for track in os.listdir(maestro_folder):
        if track.endswith(".wav"):
            track_name = track.split(".")[0]
            if track_name not in lines:
                continue
            midi_path = None
            for file in os.listdir(f"{org_folder}/{track_name}"):
                if re.match(".*(.mid|.midi)$", file):
                    midi_path = file
                    break
            song = {"id": track_name, "audio": f"{maestro_folder}/{track}",
                    "sequence": f"{org_folder}/{track_name}/{midi_path}"}
            songs.append(song)
    return songs


def parse_maestro_org(maestro_folder):
    file = open("MAESTRO_SUBSET.txt", "r")
    lines = file.read().splitlines()
    file.close()

    songs = []
    #for track_folder in os.listdir(maestro_folder):
    for track_folder in lines:
        if re.match("^(ORIG-MIDI|MIDI-Unprocessed|MIDI-UNPROCESSED)(.*)$", track_folder):
            midi_path = None
            for file in os.listdir(f"{maestro_folder}/{track_folder}"):
                if re.match(".*(.mid|.midi)$", file):
                    midi_path = file
                    break
            song = {"id": track_folder, "audio": f"{maestro_folder}/{track_folder}/{track_folder}.wav",
                    "sequence": f"{maestro_folder}/{track_folder}/{midi_path}"}
            songs.append(song)
    return songs


if __name__ == "__main__":
    # print(parse_urmp_org("Datasets/URMP2/Dataset"))
    # print(parse_urmp_org("Datasets/URMP/Dataset"))
    # print(parse_guitarset("../GuitarSet-SSH", "../Datasets/GuitarSet/train"))
    # print(parse_musicnet_rendered("../MusicNet-SSH", "../Datasets/MusicNet/tracks"))
    # print(parse_musicnet_org("../Datasets/MusicNet/tracks"))
    print(parse_maestro_rendered("../Maestro-SSH", "../Datasets/maestro-v3.0.0/train/"))
    print(parse_musicnet_org("../Datasets/maestro-v3.0.0/train"))
