"""
Script to generate TFRecords of data for all datasets specified in TFRECORD_DATASETS in config.py

Template structure:
{
    "id": name of the folder (str), i.e. "urmp_experiment_1"
    "type": type of dataset (str), i.e. "train", "validation, "test" etc
    "generate_folder": path to folder of generated audio_files (str)
    "original_folder": path to folder of original data (str)
    "use": use this dataset in processing (bool)
}
"""
import logging
import os
from dataset_config import *
from Dataset.tfrecord_generator import TFRecordGenerator
import dataset_parsers
import numpy as np
import time

def main():
    # Logging of which datasets are in use, skipped or ready to process
    datasets_unused = []
    datasets_to_skip = []
    datasets_to_process = []

    for dataset in TFRECORD_DATASETS:
        data_name = dataset["name"]
        data_dataset = dataset["dataset"]
        data_path = os.path.join(data_dataset, data_name)
        if dataset["use"]:
            if os.path.exists(os.path.join(TFRECORD_LOCATION, data_dataset, f"{data_name}.tfrecord")):
                datasets_to_skip.append(data_path)
            else:
                datasets_to_process.append(dataset)
        else:
            datasets_unused.append(data_path)

    print("Unused datasets: ", datasets_unused)
    print("Skipping already installed datasets: ", datasets_to_skip)
    if len(datasets_to_process) > 0:
        print("Processing missing datasets: ", datasets_to_process)
    else:
        print(f"All datasets already installed."
              f"Check data in dataset folder ({TFRECORD_LOCATION}) if something unexpected happened.")
        return
    del datasets_unused, datasets_to_skip

    for dataset in datasets_to_process:
        try:
            data_name = dataset["name"].lower()
            data_dataset = dataset["dataset"].lower()
            data_list = None
            if data_name == 'validation' or data_name == 'baseline':
                if 'maestro' in data_dataset:
                    data_list = dataset_parsers.parse_maestro_org(dataset["original_folder"])
                elif 'guitarset' in data_dataset:
                    data_list = dataset_parsers.parse_guitarset_org(dataset["original_folder"])
                elif 'urmp' in data_dataset:
                    data_list = dataset_parsers.parse_urmp_org(dataset["original_folder"])
                elif 'musicnet' in data_dataset:
                    data_list = dataset_parsers.parse_musicnet_org(dataset["original_folder"])
                else:
                    logging.info("Could not find dataset %s/%s", data_dataset, data_name)
            else:
                if 'maestro' in data_dataset:
                    data_list = dataset_parsers.parse_maestro_rendered(dataset["generate_folder"], dataset["original_folder"])
                elif 'guitarset' in data_dataset:
                    data_list = dataset_parsers.parse_guitarset_rendered(dataset["generate_folder"], dataset["original_folder"])
                elif 'urmp' in data_dataset:
                    data_list = dataset_parsers.parse_urmp_rendered(dataset["generate_folder"], dataset["original_folder"])
                elif 'musicnet' in data_dataset:
                    data_list = dataset_parsers.parse_musicnet_rendered(dataset["generate_folder"], dataset["original_folder"])
                else:
                    logging.info("Could not find dataset %s/%s", data_dataset, data_name)

            if "split" in dataset and dataset["split"] > 1:
                np_splits = np.array_split(data_list, dataset["split"])
                for i in range(len(np_splits)):
                    TFRecordGenerator.generate_tfrecord(data_dataset, f"{data_name}_p{i}", np_splits[0].tolist())
            else:
                curr_time = time.time()
                TFRecordGenerator.generate_tfrecord(data_dataset, data_name, data_list)
                print(f"Time to generate {data_dataset}_{data_name}: {time.time()-curr_time}")

        except ImportError:
            print("Something wrong occurred with dataset: ", dataset["id"])


if __name__ == "__main__":
    main()



