import re


class MetricSequencer:
    """
    Class with static methods to retrieve metrics from experiments folder
    Metrics are retrieved from evaluation file made by MT3 evaluation script
    """

    @staticmethod
    def get_header():
        """
        Get header of the print
        :return: The header of the print
        """
        return "{:>40}: {:>30} |{:>30} |{:>30} |".format("Name",
                                                         "Frame",
                                                         "Onset",
                                                         "Onset+Offset")

    @staticmethod
    def get_subheader():
        """
        Get subheader of the print
        :return: The subheader of the print
        """
        return "{:>40}: {:>9} |{:>9} |{:>9}|{:>9} |{:>9} |{:>9}|{:>9} |{:>9} |{:>9}|".format("", "P", "R", "F1",
                                                                                             "P", "R", "F1",
                                                                                             "P", "R", "F1")

    @staticmethod
    def get_metrics(name, file, format_string):
        """
        Get the metric values for a specific dataset
        :param name: name of the dataset
        :param file: file_path of the metric file
        :param format_string: The format string to use
        :return:
        """
        with open(file, 'rb') as f:
            file_data = f.read().decode('utf-8')
        return format_string.format(
            name,
            MetricSequencer.get_re_match("Frame Precision", file_data),
            MetricSequencer.get_re_match("Frame Recall", file_data),
            MetricSequencer.get_re_match("Frame F1", file_data),
            MetricSequencer.get_re_match("Onset precision", file_data),
            MetricSequencer.get_re_match("Onset recall", file_data),
            MetricSequencer.get_re_match("Onset F1", file_data),
            MetricSequencer.get_re_match("Onset \+ offset precision", file_data),
            MetricSequencer.get_re_match("Onset \+ offset recall", file_data),
            MetricSequencer.get_re_match("Onset \+ offset F1", file_data))

    @staticmethod
    def get_presentation_format_string():
        """
        Get format string for regular printing
        :return: format string for regular printing
        """
        return "{:>40}: {:>9.2f} |{:>9.2f} |{:>9.2f}|{:>9.2f} |{:>9.2f} |{:>9.2f}|{:>9.2f} |{:>9.2f} |{:>9.2f}|"

    @staticmethod
    def get_latex_format_string():
        """
        Get format string for latex printing
        :return: format string for latex printing
        """
        return "{:>40}: {:.2f} & {:.2f} & {:.2f} & {:.2f} & {:.2f} & {:.2f} & {:.2f} & {:.2f} & {:.2f} \\\\"

    @staticmethod
    def get_re_match(metric, file_data):
        """
        Get regular expression match
        :param metric: Metric to find
        :param file_data: data from evaluation file
        :return: The metric found
        """
        m = re.match(r"^.*'{}':([^,]+),".format(metric), file_data)
        if m:
            try:
                return float(m.group(1))
            except ValueError:
                return None
        else:
            return None


if __name__ == '__main__':
    import os

    EXPERIMENTS_PATH = '/cluster/home/fredrifw/MT3/experiments/'
    print(MetricSequencer.get_header())
    print(MetricSequencer.get_subheader())
    for directory in os.listdir(EXPERIMENTS_PATH):
        file_path = os.path.join(EXPERIMENTS_PATH, directory, 'eval_metrics.out')
        print(file_path)
        if os.path.isfile(file_path):
            print(MetricSequencer.get_metrics(directory, file_path, MetricSequencer.get_presentation_format_string()))
            print(MetricSequencer.get_metrics(directory, file_path, MetricSequencer.get_latex_format_string()))