from mido import MidiFile, MetaMessage, Message, merge_tracks, MidiTrack
import os
import numpy as np
import librosa
import pretty_midi
from pydub import AudioSegment
from tqdm import tqdm
import re
import soundfile as sf
from copy import deepcopy


def rshift_midi(midi_path, output, bars=1):
    """
    Reads a midi-file from the specified path, then pads it rightwards so that it starts the specified number of bars
    later than originally. The new (right shifted) midi is saved at the specified output location.
    :param midi_path: The path of the midi
    :param bars: Number of bars to right shift
    :param output: Where to save the shifted midi
    :return: None
    """
    mid = MidiFile(midi_path, clip=True)
    track = mid.tracks[0]
    for i in range(len(track)):
        if i != len(track) - 1:
            # Shit start length of latest MetaMessage
            if type(track[i - 1]) == MetaMessage and type(track[i]) == Message:
                # Tpb * 4 is on bar
                track[i].time += int(mid.ticks_per_beat * 4 * bars)
    mid.save(filename=f"{output}")
    # return mid


def _zero_pad_audio(audio, zeroes):
    """
    Zero pads an numpy audio array (horizontally) with the given number of zeroes across all channels
    :param audio: The audio to pad
    :param zeroes: The number of zeroes to pad
    :return: The zero padded numpy audio
    """
    new_audio = []
    for i in range(len(audio)):
        new_channel = np.hstack((audio[i], np.array([0 for _ in range(zeroes)])))
        new_audio.append(new_channel)
    return np.array(new_audio)


def _merge_audio(audio):
    """
    Takes a list of audio and merges them together.
    Audio is padded in such a way that the output audio has the same length as the longest of the audio files.
    :param audio: A list of audio signals (as numpy arrays)
    :return: A new audio signal where all the original signal
    """
    # Find length of longest audio signal, so that rest can be padded
    longest_audio_len = 0
    for a in audio:
        if a.shape[1] > longest_audio_len:
            longest_audio_len = a.shape[1]

    output_audio = audio[0]
    output_audio = _zero_pad_audio(output_audio, longest_audio_len - output_audio.shape[1])
    for i in audio[1:]:
        i = _zero_pad_audio(i, longest_audio_len - i.shape[1])
        output_audio += i
    return output_audio


def _load_audio_file(audio_path, sample_rate, duration=None):
    """
    Loads an audio file into a numpy array.
    Dimensions are (channels, sr * duration in s)
    :param audio_path: The path of the audio
    :param duration: Duration of audio to use. None if whole audio.
    :return: sig for audio
    """
    sig, rate = librosa.load(audio_path, duration=duration, mono=True, sr=sample_rate)
    assert (rate == sample_rate)
    if len(sig.shape) == 1:
        return np.array([sig])
    return sig


def merge_midi(midis, slice_index=0):
    """
    Merges a list of mido MidiFiles.
    All tracks are used from the first midi-file,
    and then everything from slice_index: is added from the rest of the tracks.
    :param midis: The list of mido MidiFiles
    :param slice_index: Start index of tracks to keep in midis[1:]
    :return: The merged midi file
    """
    mid1 = midis[0]
    merged_mid = MidiFile()
    # Copy time resolution. (If mid1 and mid2 have different
    # ticks_per_beat you will get timing problems.)
    merged_mid.ticks_per_beat = mid1.ticks_per_beat
    # Merge
    for midi in midis:
        merged_mid.tracks += midi.tracks[slice_index:]
    return merged_mid


def flac_to_wav(path):
    for root, _, files in os.walk(path):
        for file in tqdm(files):
            if file.endswith(".flac"):
                os.system(f"sox {os.path.join(root, file)} {os.path.join(root, file.split('.')[0])}.wav")
                os.remove(os.path.join(root, file))


def merge_midi_tracks(midi):
    """
    Merges all midi-tracks in the given MidiFile into one track
    :param midi: The MidiFile
    :return:
    """
    new_tracks = []
    channel_to_use = None
    for track in midi.tracks:
        new_track = MidiTrack()
        for message in track:
            new_message = deepcopy(message)
            if channel_to_use is None and not type(message) == MetaMessage:
                channel_to_use = message.channel
            if type(new_message) == Message:
                new_message.channel = channel_to_use
            new_track.append(new_message)
        new_tracks.append(new_track)
    merged = merge_tracks(new_tracks)
    midi.tracks = [merged]
    return midi


def merge_midi_tracks_in_folder(folder):
    for root, _, files in os.walk(folder):
        for file in files:
            if file.endswith(".mid") or file.endswith(".midi"):
                midi = merge_midi_tracks(MidiFile(os.path.join(root, file)))
                midi.save(os.path.join(root, file))


def jams_to_midi(jam, q=0):
    # q = 1: with pitch bend. q = 0: without pitch bend.
    midi = pretty_midi.PrettyMIDI()
    annos = jam.search(namespace='note_midi')
    if len(annos) == 0:
        annos = jam.search(namespace='pitch_midi')
    for anno in annos:
        midi_ch = pretty_midi.Instrument(program=25)
        for note in anno:
            pitch = int(round(note.value))
            bend_amount = int(round((note.value - pitch) * 4096))
            st = note.time
            dur = note.duration
            n = pretty_midi.Note(
                velocity=100 + np.random.choice(range(-5, 5)),
                pitch=pitch, start=st,
                end=st + dur
            )
            pb = pretty_midi.PitchBend(pitch=bend_amount * q, time=st)
            midi_ch.notes.append(n)
            midi_ch.pitch_bends.append(pb)
        if len(midi_ch.notes) != 0:
            midi.instruments.append(midi_ch)
    return midi


def match_target_amplitude(sound, target_dBFS):
    change_in_dBFS = target_dBFS - sound.dBFS
    return sound.apply_gain(change_in_dBFS)


def match_target_amplitude_of_files(target, org_folder, tag):
    sound = AudioSegment.from_file(target)
    org_track = None
    for f in os.listdir(org_folder):
        if re.match(f"(.*){tag}(.*).wav", f):
            org_track = f"{org_folder}/{f}"
    if not org_track:
        for f in os.listdir(org_folder):
            if re.match("(.*).wav", f):
                if org_track:
                    raise EnvironmentError("Unexpected number of files to match..")
                org_track = f"{org_folder}/{f}"
    original = AudioSegment.from_file(org_track)
    normalize_sound = match_target_amplitude(sound, original.dBFS)
    normalize_sound.export(target, format="wav")


def sftp_match_target_amplitude_of_files(target, org_folder, tag, sftp_client):
    sound = AudioSegment.from_file(target)
    org_track = None
    for f in sftp_client.listdir(org_folder):
        if re.match(f"(.*){tag}(.*).wav", f):
            org_track = f"{org_folder}/{f}"
    if not org_track:
        for f in sftp_client.listdir(org_folder):
            if re.match("(.*).wav", f):
                if org_track:
                    raise EnvironmentError("Unexpected number of files to match..")
                org_track = f"{org_folder}/{f}"
    b_org = sftp_client.open(org_track)
    b_org.prefetch()
    b_org = b_org.read()
    with open("temporary_audio_from_utils.wav", "wb") as f:
        f.write(b_org)
    os.system(f"sox temporary_audio_from_utils.wav -b 32 temporary_sox.wav")
    os.remove("temporary_audio_from_utils.wav")
    os.rename("temporary_sox.wav", "temporary_audio_from_utils.wav")
    original = AudioSegment.from_file("temporary_audio_from_utils.wav")
    print(original.dBFS)
    normalized_sound = match_target_amplitude(sound, original.dBFS)
    normalized_sound.export(target, format="wav")



def normalize_rendered_audio(rendered_folder, org_folder, org_wav_regex):
    for track in tqdm(os.listdir(rendered_folder)):
        track_name = track.replace("_hex", "")
        sound = AudioSegment.from_file(f"{rendered_folder}/{track}")
        original_name = None
        for ot in os.listdir(rendered_folder):
            if re.match(org_wav_regex, ot):
                original_name = ot
                break
        if not original_name:
            raise FileNotFoundError(f"No matching original file found for {track_name}")
        original = AudioSegment.from_file(f"{org_folder}/{track_name}/{original_name}")
        normalized_sound = match_target_amplitude(sound, original.dBFS)
        normalized_sound.export(f"{rendered_folder}/{track}", format="wav")


if __name__ == "__main__":
    pass


